# Copyright 2017, 2018, 2020 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from reahl.component.shelltools import Command, Executable, CompositeCommand
from reahl.workstation.xprasupport import VagrantMachine
from reahl.workstation.dockersupport import DockerContainer

class Ngrok(CompositeCommand):
    """Manipulates ngrok on the local machine."""
    keyword = 'ngrok'

    @property
    def commands(self):
        return [NgrokStart]


class NgrokStart(Command):
    """Starts ngrok to open a tunnel to a local / vagrant or other machine reachable via ssh."""
    keyword = 'start'
    def assemble(self):
        super().assemble()
        location_group = self.parser.add_mutually_exclusive_group(required=True)
        location_group.add_argument('-l', '--local', action='store_true', dest='local', help='exposes the local machine ssh port')
        location_group.add_argument('-V', '--vagrant', nargs='?', const='default', default=False, help='exposes the given vagrant machine')
        location_group.add_argument('-D', '--docker', nargs='?', const='reahl', default=False, help='exposes the given docker container')
        location_group.add_argument('-s', '--ssh', default=None, help='exposes the given remote machine')
        location_group.add_argument('-n', '--named-tunnel', nargs='*', default=None, help='starts the given named tunnels')
        self.parser.add_argument('-p', '--port',  default='22', help='the port to expose')
        self.parser.add_argument('-r', '--region',  default='eu', help='ngrok region to connect to')
        self.parser.add_argument('-P', '--path',  default='~/bin', help='where to find ngrok')

    def execute(self, args):
        super().execute(args)

        env_with_path = dict(os.environ)
        env_with_path['PATH'] += '%s%s' % (os.pathsep, os.path.expanduser(args.path))
        if args.named_tunnel:
            return Executable('ngrok', verbose=True).check_call(['start', '--region=%s' % args.region]+args.named_tunnel, env=env_with_path)
        else:
            if args.vagrant:
                vagrant_ssh_config = VagrantMachine(args.vagrant).get_ssh_config()
                hostname = vagrant_ssh_config['HostName']
                port = vagrant_ssh_config['Port']
            elif args.docker:
                docker_container = DockerContainer(args.docker)
                hostname = docker_container.ip_address
                port = '22'
            elif args.ssh:
                hostname = args.ssh
                port = args.port
            else:
                hostname = None
                port = args.port

            hostname_port = ':'.join([i for i in [hostname, port] if i])
            return Executable('ngrok', verbose=True).check_call(['tcp', '--region=%s' % args.region, hostname_port], env=env_with_path)




# Copyright 2013-2020 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys
import subprocess
import pkg_resources
import os
import shutil
import glob
import re
import xml.etree.ElementTree


def ask(prompt):
    answer = None
    while answer is None:
        response = input(prompt)
        if response.lower() in ('y', 'yes'):
            answer = True
        if response.lower() in ('n', 'no'):
            answer = False
    return answer

def rm_f(f):
    if os.path.isfile(f):
        os.remove(f)

def read_env_variable(variable, error_message):
    try:
      return os.environ[variable]
    except KeyError as e:
      raise AssertionError(error_message)

def clean_virtual_env(virtual_env):
    for f in glob.iglob(os.path.join(virtual_env, 'lib', 'python*', 'site-packages', '*egg-link')):
        rm_f(f)

    for f in glob.iglob(os.path.join(virtual_env, 'lib', 'python*', 'site-packages', 'easy-install.pth')):
        rm_f(f)

def clean_workspace(reahl_workspace):
    rm_f(os.path.join(reahl_workspace, '.reahlworkspace', 'workspace.projects'))
    rm_f(os.path.join(reahl_workspace, '.reahlworkspace', 'current.selection'))

def clean_egg_info_dirs():
    for current_directory, directories, files in os.walk(os.getcwd()):
        for d in directories:
            if d.endswith('egg-info') and not d in ['reahl_dev.egg-info']:
                shutil.rmtree(os.path.join(current_directory, d))

def remove_versions_from_requirements(requires_file):
    with open(requires_file, 'r') as input_file:
        lines = input_file.readlines()
    with open(requires_file, 'w') as output_file:
        for line in lines:
            version_stripped_line = re.match('([\w-]+)', line).group(0)
            output_file.write(version_stripped_line)
            output_file.write('\n')

def egg_dirs_for(project_dirs):
    for project_dir in project_dirs:
        egg_info = '%s.egg-info' % project_dir.replace('-', '_')
        yield os.path.join(os.getcwd(), project_dir, egg_info)


def fake_distributions_into_existence(project_dirs):
    for egg_dir in egg_dirs_for(project_dirs):
        if not os.path.exists(egg_dir):
            os.mkdir(egg_dir)


def find_all_prerequisits_for(project_dirs):
    prerequisites = {'tox>=3.14,<3.14.999'} # Nothing depends on tox, but we use it in .travis.yml so it forms part of our basic infrastructure
    for project_dir in project_dirs:
        prerequisites.update(parse_prerequisites_from(os.path.join(os.getcwd(), project_dir, '.reahlproject')))
    return prerequisites


def find_missing_prerequisites(project_dirs):
    missing = set()
    for i in find_all_prerequisits_for(project_dirs):
        try:
            pkg_resources.require(i)
        except (pkg_resources.DistributionNotFound, pkg_resources.VersionConflict):
            missing.add(i.strip())
    return list(missing)


def install_with_pip(package_list, upgrade=False):
    args = ['-U'] if upgrade else []
    return subprocess.call([sys.executable, '-m', 'pip', 'install'] + args + package_list)


def install_prerequisites(missing):
    print('----------------------------------------------------------------------------------')
    print('reahl-dev and its dependencies depend on eggs that are not part of reahl.')
    print('You need these in order to run this script.')
    print('Some of these are not installed on your system.')
    print('The missing eggs are:')
    print('')
    print('  '+(' '.join(missing)))
    print('')
    print('I will now go ahead and "pip install" the eggs listed above')
    print('')
    return install_with_pip(missing, upgrade=True) == 0


def get_common_version():
    def get_debian_changelog_src():
        src_lines = []
        src_line = False
        with open('reahl-dev/reahl/dev/devdomain.py', 'r') as f:
            for line in f:
                if line.startswith('# bootstrap.py-begin'):
                    src_line = True
                if line.startswith('# bootstrap.py-end'):
                    src_line = False
                if src_line:
                    src_lines.append(line)
        return '\n'.join(src_lines)
    tmp_locals = {}
    exec(get_debian_changelog_src(), {'re': re}, tmp_locals)
    DebianChangelog = tmp_locals['DebianChangelog']

    return '.'.join(DebianChangelog('debian/changelog').version.split('.')[:2])


def make_core_projects_importable(core_project_dirs):
    for d in core_project_dirs:
        project_path = os.path.join(os.getcwd(), d)
        pkg_resources.working_set.add_entry(project_path)
        sys.path.append(project_path)
    common_version = get_common_version()
    for egg_dir in egg_dirs_for(core_project_dirs):
        pkg_filename = os.path.join(egg_dir, 'PKG-INFO')
        if not os.path.exists(pkg_filename):
            with open(pkg_filename, 'w') as pkg_file:
                pkg_file.write('Version: %s\n' % common_version)

def bootstrap_workspace(workspace_dir, core_project_dirs):
    make_core_projects_importable(core_project_dirs)
    from reahl.dev.devdomain import Project, Workspace
    workspace = Workspace(workspace_dir)
    for project_dir in core_project_dirs:
        Project.from_file(workspace, os.path.join(os.getcwd(), project_dir))

    core_projects = [Project.from_file(workspace, os.path.join(os.getcwd(), project_dir)) for project_dir in core_project_dirs]
    return workspace, core_projects

def run_setup(workspace, projects, uninstall=False):
    command = ['develop', '-N']
    if uninstall:
        command.append('--uninstall')

    for project in projects:
        project.setup(command)

def find_missing_dependencies(workspace):
    missing = set()
    for project in workspace.selection:
        missing.update(project.list_missing_dependencies(for_development=True))
    return list(missing)


def print_final_message(success=True):
    debs_needed_to_compile_python = ['python-virtualenv', 'python-dev', 'python3-dev', 'gcc', 'cython', 'libxml2-dev', 'libxslt-dev', 'libsqlite3-0', 
                                    'sqlite3', 'postgresql-server-dev-all', 'zlib1g-dev', 'libjpeg62-dev', 'libfreetype6-dev', 'liblcms2-dev', 
                                    'mysql-client', 'libmysqlclient-dev']
    general_debs_needed = ['openssh-client', 'dpkg-dev', 'firefox', 'firefox-geckodriver']

    print('')
    print('')
    if success:
      print('-- ALL DONE --------------------------------------------------------------------------')
      print('Done. All python dependencies satisfied.')
    else:
      print('-- ERROR --------------------------------------------------------------------------')
      print('Something went wrong....please check that all the dependencies listed here are installed:')
    print('')
    print('In order to develop on Reahl, you will need other (non-python) packages as well.')
    print('What these are called may differ depending on your distribution/OS.')
    print('As a hint, on ubuntu systems these are:')
    print('')
    print('  '+' '.join(general_debs_needed)) 
    print('')
    print('NOTE: If you\'re working in a virtualenv and want to pip install dependencies, you will also need:')
    print('')
    print('  '+' '.join(debs_needed_to_compile_python)) 
    print('')

def parse_prerequisites_from(dot_project_file):
    root = xml.etree.ElementTree.parse(dot_project_file).getroot()
    deps_for_common_version = root.findall('.//version[@number="%s"]//thirdpartyegg' % get_common_version())
    for node in deps_for_common_version:
        requirement_string = node.attrib['name']
        min_version = node.attrib.get('minversion', None)
        max_version = node.attrib.get('maxversion', None)
        if min_version:
            requirement_string += '>=%s' % min_version
        if max_version:
            comma = ',' if min_version else ''
            requirement_string += '%s<%s' % (comma, max_version)
        yield requirement_string

    
def ensure_script_dependencies_installed():
    missing = find_missing_prerequisites(core_project_dirs)
    if missing and install_prerequisites(missing):
        return []
    elif missing:
        return missing
    return []

def ensure_reahl_project_dependencies_installed():
    workspace, core_projects = bootstrap_workspace(reahl_workspace, core_project_dirs)
    run_setup(workspace, core_projects)
    workspace.selection = core_projects

    # For good measure, we "setup.py develop" all eggs in reahl
    workspace.refresh(False, [os.getcwd()])
    workspace.select(all_=True)
    run_setup(workspace, workspace.selection)

    missing_dependencies = find_missing_dependencies(workspace)
    if missing_dependencies:
        run_setup(workspace, workspace.selection, uninstall=True)
        if install_with_pip(list(set(missing_dependencies).union(find_all_prerequisits_for(core_project_dirs))), upgrade=False) != 0:
            print("Error trying to install one of: " + ','.join(missing_dependencies))
            return False
        run_setup(workspace, workspace.selection)
    return True


virtual_env = read_env_variable('VIRTUAL_ENV', 'You should develop on Reahl within a virtualenv.')
reahl_workspace = read_env_variable('REAHLWORKSPACE',
                    'Please set the environment variable REAHLWORKSPACE to point to a parent directory of %s' \
                          % (os.getcwd()))
reahl_dev_requires_file = os.path.join(os.getcwd(), 'reahl-dev', 'reahl_dev.egg-info', 'requires.txt')
core_project_dirs = ['reahl-component', 'reahl-stubble', 'reahl-tofu', 'reahl-dev']

clean_virtual_env(virtual_env)
clean_workspace(reahl_workspace)
clean_egg_info_dirs()

remove_versions_from_requirements(reahl_dev_requires_file)
fake_distributions_into_existence(core_project_dirs)


if "--script-dependencies" in sys.argv:
   still_missing = ensure_script_dependencies_installed()
   if still_missing:
       print('Failed to install %s - (see pip errors above)' % (' '.join(still_missing)) )
   else:
       print('Successfully installed prerequisites - please re-run with --pip-installs')
       
elif "--pip-installs" in sys.argv:
   success = ensure_reahl_project_dependencies_installed()
   print_final_message(success=success)
else:
   print('Usage: %s [--script-dependencies|--pip-installs]' % sys.argv[0])
   exit(123)



# Copyright 2015-2020 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A commandline utility to display package information installed in your environment."""

import urllib
import argparse

from bs4 import BeautifulSoup

import pip
import pkg_resources
from pkg_resources import parse_version


class PypiPackage:
    def __init__(self, package_name):
        self.package_name = package_name

    @property
    def is_installed(self):
        try:
            pkg_resources.get_distribution(package_name)
        except pkg_resources.DistributionNotFound as e:
            return False
        return True

    @property
    def is_exists_in_pypi(self):
        try:
            urllib.request.urlopen(self.get_pypi_org_url())
        except urllib.error.HTTPError as e:
            return False
        return True

    def get_pypi_org_url(self, version=None):
        if version:
            return 'https://pypi.python.org/pypi/%s' % ('%s/%s' % (self.package_name, self.installed_version_string))
        return 'https://pypi.python.org/pypi/%s' % self.package_name

    @property
    def distribution(self):
        return pkg_resources.get_distribution(self.package_name)

    @property
    def installed_version_string(self):
        return self.distribution.version

    @property
    def package_name_version(self):
        if self.is_installed:
            return '%s %s' % (self.package_name, self.installed_version_string)
        return self.package_name

    @property
    def installed_version(self):
        return self.distribution.parsed_version

    @property
    def installed_license(self):
        if self.is_installed:
            metadata_resource_name = self.distribution.PKG_INFO
            if self.distribution.has_metadata(metadata_resource_name):
                license_prefix = 'License: '
                licenses = [l[len(license_prefix):] for l in self.distribution.get_metadata_lines(metadata_resource_name) if l.startswith(license_prefix)]
                return ','.join(licenses)
            return '***No license metadata for installed package***'
        return '***Package not installed***'

    @property
    def pypi_license_for_installed_version(self):
        if self.is_exists_in_pypi:
            url_to_open = self.get_pypi_org_url(version=self.installed_version_string)
            try:
                page = urllib.request.urlopen(url_to_open)
                soup = BeautifulSoup(page)
                soup.prettify()
                for anchor in soup.findAll('a', href=True):
                    if('License' in anchor.contents[0]):
                        license_text = anchor.contents[0]
                        return license_text.split(' :: ')[-1]
                return '***No license metadata on pypi***'
            except urllib.error.HTTPError as e:
                return '***Package version %s not found in pypi (%s)***' % (self.installed_version_string, url_to_open)
        return '***Package not found in pypi***'

    def get_version_from_href(self, href):
        return href.split('/')[-1]

    def other_version_is_newer(self, other_raw_version):
        return parse_version(other_raw_version) > self.installed_version

    def available_newer_versions(self):
        newer_versions = []
        page = urllib.request.urlopen(self.get_pypi_org_url())
        soup = BeautifulSoup(page)
        soup.prettify()
        for anchor in soup.findAll('a', href=True):
            href = anchor['href']
            if href.startswith('/pypi/%s' % package_name):
                available_raw_version = self.get_version_from_href(href)
                if self.other_version_is_newer(available_raw_version):
                    newer_versions.append(available_raw_version)
        return newer_versions

    def __repr__(self):
        return '%s [%s] InstalledLic[%s] PypiLic[%s] newer versions: %s' % (self.package_name, self.installed_version_string, self.installed_license,
                                                       self.pypi_license_for_installed_version(), str(self.available_newer_versions()))


parser = argparse.ArgumentParser(description='Show meta information about installed packages.')
parser.add_argument('-l', '--installed_license', dest='show_installed_license', action='store_true', default=False,
                   help='show the license info found in the installed package metadata(if package is installed)')
parser.add_argument('-L', '--pypi_license', dest='show_pypi_license', action='store_true', default=False,
                   help='show the license info found in in pypi for the installed package')
parser.add_argument('-V', '--newer_versions', dest='show_newer_versions', action='store_true', default=False,
                   help='list newer versions on pypi than the installed version')
parser.add_argument('-p', '--package_names', dest='package_names', nargs='*', metavar='package_name',
                   help='instead of all installed packages, show the meta information only for a specified list of packages')
args = parser.parse_args()

packages = args.package_names if args.package_names else [dist.project_name for dist in pip.get_installed_distributions()]
for package_name in packages:
    try:
        pypi_package = PypiPackage(package_name)

        list_of_info_to_display = [pypi_package.package_name_version]
        if args.show_installed_license:
            list_of_info_to_display.append('  Licence(Installed): %s' % pypi_package.installed_license)
        if args.show_pypi_license:
            list_of_info_to_display.append('  Licence(Pypi): %s' % pypi_package.pypi_license_for_installed_version)
        if args.show_newer_versions:
            newer_versions = pypi_package.available_newer_versions()
            list_of_info_to_display.append('  Newer versions: %s' % ('-' if not newer_versions else ', '.join(newer_versions)))
        print('\n'.join(list_of_info_to_display))
    except:
        print ('Problem retrieving details for package name: %s' % package_name)
